import matplotlib.pyplot as plt 
import numpy as np 
import pdb
import matplotlib.gridspec as gridspec
import matplotlib
# matplotlib.rcParams.update(matplotlib.rcParamsDefault)
"""-
init
-"""
nside = 4
seeds = [10,20,30,40,50,60] 


"""------------
Stationary case
---------------"""
mode1 = 's' #'s' for stationary, 'o' for orbiting  

logz_s = np.array([])
logzerr_s = np.array([])

"""-------------------------------------------------------------------------------------------
filename format : /home/tommy/lisaenv/blip/Feb19/[mode]_n[nside]_s[seed#]/logz_rs[seed#]_[mode].txt
for nside=4, seed=10, mode=s: /home/tommy/lisaenv/blip/s_n4_s10/logz_rs10_s.txt
----------------------------------------------------------------------------------------------"""
for x in range(len(seeds)):
    file = open("/home/tommy/lisaenv/blip/Feb26/" + mode1 + "_n" + str(nside) + "_s" + str(seeds[x]) + "/logz_rs" + str(seeds[x]) + "_" + mode1 + ".txt")
    last_line = file.readlines()[-1].replace("\n",'')
    logz_s = np.append(logz_s, float("{:.8e}".format(float(last_line))))
    # logz_s = np.append(logz_s, float(last_line))


"""----------------------------------------------------------------------------------------------
filename format : /home/tommy/lisaenv/blip/Feb19/[mode]_n[nside]_s[seed#]/logzerr_rs[seed#]_[mode].txt
for nside=4, seed=10, mode=s: /home/tommy/lisaenv/blip/s_n4_s10/logzerr_rs10_s.txt
-------------------------------------------------------------------------------------------------"""
for x in range(len(seeds)):
    file = open("/home/tommy/lisaenv/blip/Feb26/" + mode1 + "_n" + str(nside) + "_s" + str(seeds[x]) + "/logzerr_rs" + str(seeds[x]) + "_" + mode1 + ".txt")
    last_line = file.readlines()[-1].replace("\n",'')
    logzerr_s = np.append(logzerr_s, float("{:.3e}".format(float(last_line))))
    # logzerr_s = np.append(logzerr_s, float(last_line))

"""----------
Orbiting case
-------------"""
mode2 = 'o'

logz_o = np.array([])
logzerr_o = np.array([])

for x in range(len(seeds)):
    file = open("/home/tommy/lisaenv/blip/Feb26/" + mode2 + "_n" + str(nside) + "_s" + str(seeds[x]) + "/logz_rs" + str(seeds[x]) + "_" + mode2 + ".txt")
    last_line = file.readlines()[-1].replace("\n",'')
    logz_o = np.append(logz_o, float("{:.8e}".format(float(last_line))))
    # logz_o = np.append(logz_o, int(last_line))


for x in range(len(seeds)):
    file = open("/home/tommy/lisaenv/blip/Feb26/" + mode2 + "_n" + str(nside) + "_s" + str(seeds[x]) + "/logzerr_rs" + str(seeds[x]) + "_" + mode2 + ".txt")
    last_line = file.readlines()[-1].replace("\n",'')
    logzerr_o = np.append(logzerr_o, float("{:.3e}".format(float(last_line))))
    # logzerr_o = np.append(logzerr_o, int(last_line))

# logz_o = np.array(logz_o, dtype=np.float128)
# logz_s = np.array(logz_s, dtype=np.float128)
# logzerr_o = np.array(logzerr_o, dtype=np.float128)
# logzerr_s = np.array(logzerr_s, dtype=np.float128)

"""
nside = 8 data
"""
nside2 = 8
logz_s2 = []
logz_o2 = []

for x in range(len(seeds)):
    files = open("/home/tommy/lisaenv/blip/Feb26/" + mode1 + "_n" + str(nside2) + "_s" + str(seeds[x]) + "/logz_rs" + str(seeds[x]) + "_" + mode1 + ".txt")
    last_line = files.readlines()[-1].replace("\n",'')
    logz_s2 = np.append(logz_s2, float("{:.8e}".format(float(last_line))))
    
for x in range(len(seeds)):
    files = open("/home/tommy/lisaenv/blip/Feb26/" + mode2 + "_n" + str(nside2) + "_s" + str(seeds[x]) + "/logz_rs" + str(seeds[x]) + "_" + mode2 + ".txt")
    last_line = files.readlines()[-1].replace("\n",'')
    logz_o2 = np.append(logz_o2, float("{:.8e}".format(float(last_line))))

"""-----
ANALYSIS
--------"""
log_bayes_factor = logz_o - logz_s
log_bayes_factor_8 = logz_o2 - logz_s2
fig = plt.figure(figsize=(10,7))

gs = gridspec.GridSpec(2,2)

plt.subplot(gs[0,0])
plt.plot(seeds, abs(log_bayes_factor), 'o', label='nside = 4')
plt.xlabel('Seed number')
plt.ylabel(r"$\left|log_{10} \ \frac{z_o}{z_s}\right|$")
plt.title("nside = 4")

plt.subplot(gs[0,1])
plt.plot(seeds, abs(log_bayes_factor_8), 'o', color='#D35400', label='nside = 8')
plt.xlabel('Seed number')
plt.ylabel(r"$\left|log_{10} \ \frac{z_o}{z_s}\right|$")
plt.title("nside = 8")

plt.subplot(gs[1,:])
plt.plot(seeds, abs(log_bayes_factor), 'o', label='nside = 4')
plt.plot(seeds, abs(log_bayes_factor_8), 'o', color='#D35400',label='nside = 8')

plt.title(" Log Bayes' Factor of Orbiting vs Stationary for nside = 4 \& 8")
plt.xlabel('Seed number')
# plt.ylabel(r"$log_{10} \ \frac{z_o}{z_s}$")
plt.ylabel(r"$\left|\log_{10} \ \frac{z_o}{z_s} \right|$ ")
# plt.ylabel("\log_{10} \ \frac{z_o}{z_s}")

# plt.subplot_tool()
plt.legend()
plt.show()

# sqrt(logzoerr^2 + logzserr^2)
"""
Data
"""
# out_file = open("bayesFactor_n4_n8.txt", 'w')
# print("nside = 4 logz_o", file=out_file)
# print(logz_o, file=out_file)
# print(file=out_file)
# print("nside = 4 logz_s", file=out_file)
# print(logz_s, file=out_file)
# print(file=out_file)
# print("nside = 4 logz_o - logz_s", file=out_file)
# print(log_bayes_factor, file=out_file)
# print(file=out_file)
# print(file=out_file)
# print("nside = 8 logz_o", file=out_file)
# print(logz_o2,file=out_file)
# print(file=out_file)
# print("nside = 8 logz_s", file=out_file)
# print(logz_s2,file=out_file)
# print(file=out_file)
# print("nside = 8 logz_o - logz_s", file=out_file)
# print(log_bayes_factor_8, file=out_file)



import numpy as np

@profile
def isgwb_mich_strain_response(self, f0):

        '''
        Calculate the detector transfer function functions to an isotropic SGWB non-polarized using basic michelson
        channels. Note that since this is the response to an isotropic background, the response function is integrated
        over sky direction and averaged over polarozation. The angular integral is a linear and rectangular in the
        cos(theta) and phi space.  Note also that f0 is (pi*L*f)/c and is input as an array. The response function is given
        for the strain of the signal rather than the power

        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)

    

        Returns
        ---------

        R1, R2 and R3   :   float
            Antenna Patterns for the given sky direction for the three channels, integrated over sky direction and averaged over polarization.
        '''

        pi_val = np.pi
        tt = np.linspace(-1, 1, 200)
        ppi = np.linspace(0, 2*pi_val, 200, endpoint=False)

        [ct, phi] = np.meshgrid(tt,ppi)
        ct = ct.astype('float16')
        phi = phi.astype('float16')
        dct = ct[0, 1] - ct[0,0]
        dphi = phi[1,0] - phi[0,0]

        npix = ct.size
        norm = np.sqrt(0.5/npix)

        ## udir is just u.r, where r is the directional vector
        udir = (np.sqrt(1-ct**2) * np.sin(phi + pi_val/6)).astype('float16')
        vdir = (np.sqrt(1-ct**2) * np.sin(phi - pi_val/6)).astype('float16')
        wdir = vdir - udir

        # Initlize arrays for the detector reponse
        R1 = np.zeros((f0.size, 2), dtype='c16')
        R2 = np.zeros((f0.size, 2), dtype='c16')
        R3 = np.zeros((f0.size, 2), dtype='c16')


        ## response function u x u : eplus
        ##  Fplus_u = (u x u):eplus

        Fplus_u   = (1/4*(1-ct**2) + 1/2*(ct**2)*(np.cos(phi))**2 - \
                        np.sqrt(3/16)*np.sin(2*phi)*(1+ct**2)  + \
                            0.5*((np.cos(phi))**2 - ct**2))
        
        Fplus_v   = (1/4*(1-ct**2) + 1/2*(ct**2)*(np.cos(phi))**2 + \
                        np.sqrt(3/16)*np.sin(2*phi)*(1+ct**2) + \
                            0.5*((np.cos(phi))**2 - ct**2))

        Fplus_w   = (1 - (1+ct**2)*(np.cos(phi))**2)

        ##  Fcross_u = 1/2(u x u)Gamma(udir, f):ecross
        Fcross_u  = - ct * (np.sin(2*phi + pi_val/3))
        Fcross_v  = - ct * (np.sin(2*phi - pi_val/3))
        Fcross_w   = ct*np.sin(2*phi)
        
        uminus_1 = np.multiply.outer(f0, (1-udir)/pi_val).astype('float16')
        vminus_1 = np.multiply.outer(f0, (1-vdir)/pi_val).astype('float16')
        wminus_1 = np.multiply.outer(f0, (1-wdir)/pi_val).astype('float16')

        uplus_1 = np.multiply.outer(f0, (1+udir)/pi_val).astype('float16')
        vplus_1 = np.multiply.outer(f0, (1+vdir)/pi_val).astype('float16')
        wplus_1 = np.multiply.outer(f0, (1+wdir)/pi_val).astype('float16')


        uminus_3 = np.multiply.outer(f0, (3-udir)).astype('float16')
        vminus_3 = np.multiply.outer(f0, (3-vdir)).astype('float16')
        wminus_3 = np.multiply.outer(f0, (3-wdir)).astype('float16')

        uplus_3 = np.multiply.outer(f0, (3+udir)).astype('float16')
        vplus_3 = np.multiply.outer(f0, (3+vdir)).astype('float16')
        wplus_3 = np.multiply.outer(f0, (3+wdir)).astype('float16')

        gammaU_plus = 1/2 * ( np.sinc(uminus_1)*np.exp(-1j*uplus_3) + np.sinc(uplus_1)*np.exp(-1j*pi_val*uplus_1) ).astype('c16')
        gammaV_plus = 1/2 * ( np.sinc(vminus_1)*np.exp(-1j*vplus_3) + np.sinc(vplus_1)*np.exp(-1j*pi_val*vplus_1) ).astype('c16')
        gammaW_plus = 1/2 * ( np.sinc(wminus_1)*np.exp(-1j*wplus_3) + np.sinc(wplus_1)*np.exp(-1j*pi_val*wplus_1) ).astype('c16')

        gammaU_minus = 1/2 * ( np.sinc(uplus_1)*np.exp(-1j*uminus_3) + np.sinc(uminus_1)*np.exp(-1j*pi_val*uminus_1) ).astype('c16')
        gammaV_minus = 1/2 * ( np.sinc(vplus_1)*np.exp(-1j*vminus_3) + np.sinc(vminus_1)*np.exp(-1j*pi_val*vminus_1) ).astype('c16')
        gammaW_minus = 1/2 * ( np.sinc(wplus_1)*np.exp(-1j*wminus_3) + np.sinc(wminus_1)*np.exp(-1j*pi_val*wminus_1) ).astype('c16')


        ## Michelson antenna patterns: Calculate Fplus
        Fplus1 = 0.5*(Fplus_u[None,:,:] * gammaU_plus  - Fplus_v[None,:,:] * gammaV_plus)
        Fplus2 = 0.5*(Fplus_w[None,:,:] * gammaW_plus  - Fplus_u[None,:,:] * gammaU_minus) * np.exp(2j * np.multiply.outer(f0, udir).astype('c16'))
        Fplus3 = 0.5*(Fplus_v[None,:,:] * gammaV_minus - Fplus_w[None,:,:] * gammaW_minus)*np.exp(2j * np.multiply.outer(f0, vdir).astype('c16'))

        ## Michelson antenna patterns: Calculate Fcross
        Fcross1 = 0.5*(Fcross_u[None,:,:] * gammaU_plus  - Fcross_v[None,:,:] * gammaV_plus).astype('c16')
        Fcross2 = 0.5*(Fcross_w[None,:,:] * gammaW_plus  - Fcross_u[None,:,:] * gammaU_minus) * np.exp(2j * np.multiply.outer(f0, udir).astype('c16'))
        Fcross3 = 0.5*(Fcross_v[None,:,:] * gammaV_minus - Fcross_w[None,:,:] * gammaW_minus) * np.exp(2j * np.multiply.outer(f0, vdir).astype('c16'))

        del gammaU_plus, gammaV_plus, gammaW_plus, gammaU_minus, gammaV_minus, gammaW_minus

        rand_plus = (np.random.normal(size=Fplus1.shape) + 1j* np.random.normal(size=Fplus1.shape)).astype('c16')
        rand_cross = (np.random.normal(size=Fplus1.shape) + 1j* np.random.normal(size=Fplus1.shape)).astype('c16')

        ## Detector response summed over polarization and integrated over sky direction
        R1[:, 0], R1[:, 1] = norm*(Fplus1*rand_plus).sum(axis=(1, 2)), norm*(Fcross1*rand_cross).sum(axis=(1, 2)) 
        R2[:, 0], R2[:, 1] = norm*(Fplus2*rand_plus).sum(axis=(1, 2)), norm*(Fcross2*rand_cross).sum(axis=(1, 2)) 
        R3[:, 0], R3[:, 1] = norm*(Fplus3*rand_plus).sum(axis=(1, 2)), norm*(Fcross3*rand_cross).sum(axis=(1, 2))

        return R1, R2, R3


import numpy as np
from scipy.special import lpmn
import healpy as hp

class orbitinglisa():

    '''
    Module containing geometry methods accounting for an orbiting satellite. The methods here include a base orbit, a single doppler channel, for 
    the three michelson channels or for the AET TDI channels and calculation of noise power spectra for various channel combinations. 
    '''

    
    def lisa_orbits(self, tsegmid):
    
        '''
        Define LISA orbital positions at the midpoint of each time integration segment using analytic MLDC orbits.
        
        Parameters
        -----------
        
        tsegmid  :  array
            A numpy array of the tsegmid for each time integration segment.
            
        Returns
        -----------
        rs1, rs2, rs3  :  array
            Arrays of satellite positions for each segment midpoint in timearray. e.g. rs1[1] is [x1,y1,z1] at t=midpoint[1]=timearray[1]+(segment length)/2. 
        '''

        ## Semimajor axis in m
        a = 1.496e11
   
        
        ## Alpha and beta phases allow for changing of initial satellite orbital phases; default initial conditions are alphaphase=betaphase=0.
        betaphase = 0
        alphaphase = 0
        
        ## Orbital angle alpha(t)
        at = (2*np.pi/31557600)*tsegmid + alphaphase
        
        ## Eccentricity. L-dependent, so needs to be altered for time-varied arm length case.
        e = self.armlength/(2*a*np.sqrt(3))
        
        ## Initialize arrays
        beta_n = (2/3)*np.pi*np.array([0,1,2])+betaphase

        ## meshgrid arrays
        Beta_n, Alpha_t = np.meshgrid(beta_n, at)

        ## Calculate inclination and positions for each satellite.
        x_n = a*np.cos(Alpha_t) + a*e*(np.sin(Alpha_t)*np.cos(Alpha_t)*np.sin(Beta_n) - (1+(np.sin(Alpha_t))**2)*np.cos(Beta_n)) 
        y_n = a*np.sin(Alpha_t) + a*e*(np.sin(Alpha_t)*np.cos(Alpha_t)*np.cos(Beta_n) - (1+(np.cos(Alpha_t))**2)*np.sin(Beta_n))
        z_n = -np.sqrt(3)*a*e*np.cos(Alpha_t - Beta_n)


        ## Construct position vectors r_n
        rs1 = np.array([x_n[:, 0],y_n[:, 0],z_n[:, 0]])
        rs2 = np.array([x_n[:, 1],y_n[:, 1],z_n[:, 1]])
        rs3 = np.array([x_n[:, 2],y_n[:, 2],z_n[:, 2]])

        return rs1, rs2, rs3
        
        
    def orbiting_doppler_response(self, f0, theta, phi, tsegmid):
        
        '''
        Calculate antenna pattern/ detector transfer functions for a GW originating in the direction of (theta, phi) for the u doppler channel of an orbiting LISA with satellite position vectors rs1, rs2, rs3. Return the detector response for + and x polarization. Note that f0 is (pi*L*f)/c and is input as an array.
        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)

        phi theta  :  float
            Sky position values. 
            
        tsegmid  :  array
            A numpy array of the midpoints for each time integration segment.
            
        rs1, rs2, rs3  :  array
            Satellite position vectors.
        

        Returns
        ---------

        Rplus, Rcross   :   float
            Plus and cross antenna Patterns for the given sky direction for each time in midpoints.
        '''
        print('Calculating detector response functions...')
        
        self.rs1, self.rs2, self.rs3 = self.lisa_orbits(tsegmid)

        ## Indices of midpoints array
        timeindices = np.arange(len(tsegmid))
        
        ## Define cos/sin(theta)
        ct = np.cos(theta)
        st = np.sqrt(1-ct**2)
        
        ## Initlize arrays for the detector reponse
        Rplus, Rcross = np.zeros((len(timeindices),f0.size), dtype=complex), np.zeros((len(timeindices),f0.size),dtype=complex)

        for ti in timeindices:
            ## Define x/y/z for each satellite at time given by timearray[ti]
            x1 = rs1[0][ti]
            y1 = rs1[1][ti]
            z1 = rs1[2][ti]
            x2 = rs2[0][ti]
            y2 = rs2[1][ti]
            z2 = rs2[2][ti]
            
            ## Add if calculating v, w:
            ## x3 = r3[0][ti]
            ## y3 = r3[1][ti]
            ## z3 = r3[2][ti]
            
            ## Define vector u at time tsegmid[ti]
            uvec = rs2[:,ti] - rs1[:,ti]
            ## Calculate arm length for the u arm
            Lu = np.sqrt(np.dot(uvec,uvec))
            ## udir is just u-hat.omega, where u-hat is the u unit vector and omega is the unit vector in the sky direction of the GW signal
            udir = ((x2-x1)/Lu)*np.cos(phi)*st + ((y2-y1)/Lu)*np.sin(phi)*st + ((z2-z1)/Lu)*ct
            
            ## Calculate 1/2(u x u):eplus
            Pcontract = 1/2*((((x2-x1)/Lu)*np.sin(phi)-((y2-y1)/Lu)*np.cos(phi))**2 - \
                             (((x2-x1)/Lu)*np.cos(phi)*ct+((y2-y1)/Lu)*np.sin(phi)*ct- \
                              ((z2-z1)/Lu)*st)**2)
             ## Calculate 1/2(u x u):ecross
            Ccontract = ((((x2-x1)/Lu)*np.sin(phi)-((y2-y1)/Lu)*np.cos(phi)) * \
                          (((x2-x1)/Lu)*np.cos(phi)*ct+((y2-y1)/Lu)*np.sin(phi)*ct- \
                           ((z2-z1)/Lu)*st))
    
            # Calculate the detector response for each frequency
            for ii in range(0, f0.size):
                # Calculate GW transfer function for the michelson channels
                gammaU = 1/2 * (np.sinc(f0[ii]*(1-udir)/np.pi)*np.exp(-1j*f0[ii]*(3+udir)) + \
                                    np.sinc(f0[ii]*(1+udir)/np.pi)*np.exp(-1j*f0[ii]*(1+udir)))
        
        
                ## Michelson Channel Antenna patterns for + pol: Rplus = 1/2(u x u)Gamma(udir, f):eplus
        
                Rplus[ti][ii] = Pcontract*gammaU
                
                ## Michelson Channel Antenna patterns for x pol: Rcross = 1/2(u x u)Gamma(udir, f):ecross
        
                Rcross[ti][ii] = Ccontract*gammaU
        
        return Rplus, Rcross


    def orbiting_michelson_response(self, f0, theta, phi, tsegmid): 

        '''
        Calculate Antenna pattern/ detector transfer function for a GW originating in the direction of (theta, phi) at a given time for the three Michelson channels of an orbiting LISA. Return the detector response for + and x polarization. Note that f0 is (pi*L*f)/c and is input as an array
        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)

        phi theta  : float
            Sky position values. 
            
        rs1, rs2, rs3  :  arrays
            Satellite position vectors.
            
        tsegmid  :  array
            A numpy array of the midpoints for each time integration segment.
    

        Returns
        ---------

        R1plus, R1cross, R2plus, R2cross, R3plus, R3cross   :   arrays
            Plus and cross antenna Patterns for the given sky direction for the three channels for each time in midpoints.
        '''
        print('Calculating detector response functions...')
        
        self.rs1, self.rs2, self.rs3 = self.lisa_orbits(tsegmid)

        ## Indices of midpoints array
        timeindices = np.arange(len(tsegmid))
        
        ## Define cos/sin(theta)
        ct = np.cos(theta)
        st = np.sqrt(1-ct**2)
        
        for ti in timeindices:
            ## Define x/y/z for each satellite at time given by tsegmid[ti]
            x1 = rs1[0][ti]
            y1 = rs1[1][ti]
            z1 = rs1[2][ti]
            x2 = rs2[0][ti]
            y2 = rs2[1][ti]
            z2 = rs2[2][ti]
            x3 = rs3[0][ti]
            y3 = rs3[1][ti]
            z3 = rs3[2][ti]
            
            ## Define vector u at time timearray[ti]
            uvec = rs2[:,ti] - rs1[:,ti]
            vvec = rs3[:,ti] - rs1[:,ti]
            wvec = rs3[:,ti] - rs2[:,ti]
    
            ## Calculate arm lengths
            Lu = np.sqrt(np.dot(uvec,uvec))
            Lv = np.sqrt(np.dot(vvec,vvec))
            Lw = np.sqrt(np.dot(wvec,wvec))
         
            ## udir is just u-hat.omega, where u-hat is the u unit vector and omega is the unit vector in the sky direction of the GW signal
            udir = ((x2-x1)/Lu)*np.cos(phi)*st + ((y2-y1)/Lu)*np.sin(phi)*st + ((z2-z1)/Lu)*ct
            vdir = ((x3-x1)/Lv)*np.cos(phi)*st + ((y3-y1)/Lv)*np.sin(phi)*st + ((z3-z1)/Lv)*ct
            wdir = ((x3-x2)/Lw)*np.cos(phi)*st + ((y3-y2)/Lw)*np.sin(phi)*st + ((z3-z2)/Lw)*ct
            
            ## Calculate 1/2(u x u):eplus
            Pcontract_u = 1/2*((((x2-x1)/Lu)*np.sin(phi)-((y2-y1)/Lu)*np.cos(phi))**2 - \
                             (((x2-x1)/Lu)*np.cos(phi)*ct+((y2-y1)/Lu)*np.sin(phi)*ct-((z2-z1)/Lu)*st)**2)
            Pcontract_v = 1/2*((((x3-x1)/Lv)*np.sin(phi)-((y3-y1)/Lv)*np.cos(phi))**2 - \
                             (((x3-x1)/Lv)*np.cos(phi)*ct+((y3-y1)/Lv)*np.sin(phi)*ct-((z3-z1)/Lv)*st)**2)
            Pcontract_w = 1/2*((((x3-x2)/Lw)*np.sin(phi)-((y3-y2)/Lw)*np.cos(phi))**2 - \
                             (((x3-x2)/Lw)*np.cos(phi)*ct+((y3-y2)/Lw)*np.sin(phi)*ct-((z3-z2)/Lw)*st)**2)
            
            ## Calculate 1/2(u x u):ecross
            Ccontract_u = (((x2-x1)/Lu)*np.sin(phi)-((y2-y1)/Lu)*np.cos(phi)) * \
                            (((x2-x1)/Lu)*np.cos(phi)*ct+((y2-y1)/Lu)*np.sin(phi)*ct-((z2-z1)/Lu)*st)
            
            Ccontract_v = (((x3-x1)/Lv)*np.sin(phi)-((y3-y1)/Lv)*np.cos(phi)) * \
                            (((x3-x1)/Lv)*np.cos(phi)*ct+((y3-y1)/Lv)*np.sin(phi)*ct-((z3-z1)/Lv)*st)
            
            Ccontract_w = (((x3-x2)/Lw)*np.sin(phi)-((x3-x2)/Lw)*np.cos(phi)) * \
                            (((x3-x2)/Lw)*np.cos(phi)*ct+((y3-y2)/Lw)*np.sin(phi)*ct-((z3-z2)/Lw)*st)


            ## Calculate the detector response for each frequency
            for ii in range(0, f0.size):
    
                ## Calculate GW transfer function for the michelson channels
                gammaU_p    =    1/2 * (np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(3 + udir)) + \
                                        np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(1 + udir)))
                gammaU_m    =    1/2 * (np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(3 - udir)) + \
                                        np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(1 - udir)))
                
                gammaV_p    =    1/2 * (np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + vdir)) + \
                                        np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(1+vdir)))
                gammaV_m    =    1/2 * (np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - vdir)) + \
                                        np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - vdir)))
                
                gammaW_p    =    1/2 * (np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + wdir)) + \
                                        np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 + wdir)))
                gammaW_m    =    1/2 * (np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - wdir)) + \
                                        np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - wdir)))
                ## Michelson Channel Antenna patterns for + pol
                ## Fplus_u = 1/2(u x u)Gamma(udir, f):eplus
    
                Fplus_u_p   = Pcontract_u*gammaU_p
                Fplus_u_m   = Pcontract_u*gammaU_m
                Fplus_v_p   = Pcontract_v*gammaV_p
                Fplus_v_m   = Pcontract_v*gammaV_m
                Fplus_w_p   = Pcontract_w*gammaW_p
                Fplus_w_m   = Pcontract_w*gammaW_m
    
                ## Michelson Channel Antenna patterns for x pol
                ## Fcross_u = 1/2(u x u)Gamma(udir, f):ecross
                Fcross_u_p  = Ccontract_u*gammaU_p
                Fcross_u_m  = Ccontract_u*gammaU_m
                Fcross_v_p  = Ccontract_v*gammaV_p
                Fcross_v_m  = Ccontract_v*gammaV_m
                Fcross_w_p  = Ccontract_w*gammaW_p
                Fcross_w_m  = Ccontract_w*gammaW_m
    
    
                ## First Michelson antenna patterns
                ## Calculate Fplus
                R1plus = (Fplus_u_p - Fplus_v_p)
                R2plus = (Fplus_w_p - Fplus_u_m)
                R3plus = (Fplus_v_m - Fplus_w_m)
    
                ## Calculate Fcross
                R1cross = (Fcross_u_p - Fcross_v_p)
                R2cross = (Fcross_w_p - Fcross_u_m)
                R3cross = (Fcross_v_m - Fcross_w_m)
        

        return R1plus, R1cross, R2plus, R2cross, R3plus, R3cross

    def orbiting_aet_response(self, f0, theta, phi, tsegmid): 



        '''
        Calculate Antenna pattern/ detector transfer functions for a GW originating in the direction of (theta, phi) for the A, E and T TDI channels of an orbiting LISA. Return the detector responses for + and x polarization. Note that f0 is (pi*L*f)/c and is input as an array
        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)

        phi theta  : float
            Sky position values. 
                   
        tsegmid  :  array
            A numpy array of the midpoints for each time integration segment.
            
        rs1, rs2, rs3  :  array
            Satellite position vectors.


        Returns
        ---------

        RAplus, RAcross, REplus, REcross, RTplus, RTcross   :   arrays
            Plus and cross antenna Patterns for the given sky direction for the three channels for each time in midpoints.
        '''

        self.rs1, self.rs2, self.rs3 = self.lisa_orbits(tsegmid)

        R1plus, R1cross, R2plus, R2cross, R3plus, R3cross  = self.michelson_response(f0, theta, phi, tsegmid)
        

        ## Calculate antenna patterns for the A, E and T channels
        RAplus = (2/3)*np.sin(2*f0)*(2*R1plus - R2plus - R3plus)
        REplus = (2/np.sqrt(3))*np.sin(2*f0)*(R3plus - R2plus)
        RTplus = (1/3)*np.sin(2*f0)*(R1plus + R3plus + R2plus)

        RAcross = (2/3)*np.sin(2*f0)*(2*R1cross - R2cross - R3cross)
        REcross = (2/np.sqrt(3))*np.sin(2*f0)*(R3cross - R2cross)
        RTcross = (1/3)*np.sin(2*f0)*(R1cross + R3cross + R2cross)

        return RAplus, RAcross, REplus, REcross, RTplus, RTcross
    
    def isgwb_omich_response(self, f0, tsegmid):

        '''
        Calcualte the Antenna pattern/ detector transfer function functions for an orbiting LISA to an isotropic SGWB using basic michelson
        channels. Note that since this is the response to an isotropic background, the response function is integrated
        over sky direction and averaged over polarization. The angular integral is a linear and rectangular in the
        cos(theta) and phi space.  Note also that f0 is (pi*L*f)/c and is input as an array

        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)
        
        tsegmid  :  array
            A numpy array of the midpoints for each time integration segment.
            
        rs1, rs2, rs3  :  array
            Satellite position vectors.
    

        Returns
        ---------

        R1, R2 and R3   :   float
            Antenna Patterns for the given sky direction for the three channels, integrated over sky direction and averaged over polarization, evaluated at the midpoint of each time segment.
        '''
                
        self.rs1, self.rs2, self.rs3 = self.lisa_orbits(tsegmid)

        ## Setup healpix map
        npix = hp.nside2npix(self.params['nside'])

        # Array of pixel indices
        pix_idx  = np.arange(npix)

        #Angular coordinates of pixel indcides
        theta, phi = hp.pix2ang(self.params['nside'], pix_idx)

        # Take cosine.
        ctheta = np.cos(theta)
        stheta = np.sqrt(1-ctheta**2)

        # Area of each pixel in sq.radians
        dOmega = hp.pixelfunc.nside2pixarea(self.params['nside'])

        ## Indices of midpoints array
        timeindices = np.arange(len(tsegmid))

        # Initlize arrays for the detector reponse
        R1 = np.zeros((len(timeindices),f0.size))
        R2 = np.zeros((len(timeindices),f0.size))
        R3 = np.zeros((len(timeindices),f0.size))

        ## arm vectors
        uvec = self.rs2 - self.rs1
        vvec = self.rs3 - self.rs1
        wvec = self.rs3 - self.rs2

        ## Calculate arm lengths
        Lu = np.sqrt(np.sum(uvec*uvec, axis=0))
        Lv = np.sqrt(np.sum(vvec*vvec, axis=0))
        Lw = np.sqrt(np.sum(wvec*wvec, axis=0))

        ## Define x/y/z for each satellite at all times
        x1, y1, z1 = self.rs1[0, :], self.rs1[1, :], self.rs1[2, :] 
        x2, y2, z2 = self.rs2[0, :], self.rs2[1, :], self.rs2[2, :] 
        x3, y3, z3 = self.rs3[0, :], self.rs3[1, :], self.rs3[2, :] 
        
        udir = np.tensordot((x2-x1)/Lu, np.cos(phi)*stheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z2-z1)/Lu, ctheta, axes = 0)

        vdir = np.tensordot((x3-x1)/Lv, np.cos(phi)*stheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z3-z1)/Lv, ctheta, axes = 0)

        wdir = np.tensordot((x3-x2)/Lw, np.cos(phi)*stheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z3-z2)/Lw, ctheta, axes = 0)


        ## Calculate 1/2(u x u):eplus
        Pcontract_u = 1/2*( (np.tensordot((x2-x1)/Lu, np.sin(phi), axes=0) - np.tensordot((y2-y1)/Lu, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x2-x1)/Lu, np.cos(phi)*ctheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z2-z1)/Lu, stheta, axes=0))**2  )

        Pcontract_v = 1/2*( (np.tensordot((x3-x1)/Lv, np.sin(phi), axes=0) - np.tensordot((y3-y1)/Lv, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x3-x1)/Lv, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z3-z1)/Lv, stheta, axes=0))**2  )

        Pcontract_w = 1/2*( (np.tensordot((x3-x2)/Lw, np.sin(phi), axes=0) - np.tensordot((y3-y2)/Lw, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x3-x2)/Lw, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z3-z2)/Lw, stheta, axes=0))**2  )                           


        Ccontract_u = (np.tensordot((x2-x1)/Lu, np.sin(phi), axes=0) - np.tensordot((y2-y1)/Lu, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x2-x1)/Lu, np.cos(phi)*ctheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z2-z1)/Lu, stheta, axes=0) )  

        Ccontract_v = (np.tensordot((x3-x1)/Lv, np.sin(phi), axes=0) - np.tensordot((y3-y1)/Lv, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x3-x1)/Lv, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z3-z1)/Lv, stheta, axes=0) ) 

        Ccontract_w = (np.tensordot((x3-x2)/Lw, np.sin(phi), axes=0) - np.tensordot((y3-y2)/Lw, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x3-x2)/Lw, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z3-z2)/Lw, stheta, axes=0) ) 



        # Calculate the detector response for each frequency
        for ii in range(0, f0.size):
            # Calculate GW transfer function for the michelson channels
            gammaU_p    =    1/2 * (np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(3 + udir)) + \
                                        np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(1 + udir)))
            gammaU_m    =    1/2 * (np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(3 - udir)) + \
                                        np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(1 - udir)))
                
            gammaV_p    =    1/2 * (np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + vdir)) + \
                                        np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(1+vdir)))
            gammaV_m    =    1/2 * (np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - vdir)) + \
                                        np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - vdir)))
                
            gammaW_p    =    1/2 * (np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + wdir)) + \
                                        np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 + wdir)))
            gammaW_m    =    1/2 * (np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - wdir)) + \
                                        np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - wdir)))
            
            ## Michelson Channel Antenna patterns for + pol
            ## Fplus_u = 1/2(u x u)Gamma(udir, f):eplus
            Fplus_u_p   = Pcontract_u*gammaU_p
            Fplus_u_m   = Pcontract_u*gammaU_m
            Fplus_v_p   = Pcontract_v*gammaV_p
            Fplus_v_m   = Pcontract_v*gammaV_m
            Fplus_w_p   = Pcontract_w*gammaW_p
            Fplus_w_m   = Pcontract_w*gammaW_m
    
            ## Michelson Channel Antenna patterns for x pol
            ## Fcross_u = 1/2(u x u)Gamma(udir, f):ecross
            Fcross_u_p  = Ccontract_u*gammaU_p
            Fcross_u_m  = Ccontract_u*gammaU_m
            Fcross_v_p  = Ccontract_v*gammaV_p
            Fcross_v_m  = Ccontract_v*gammaV_m
            Fcross_w_p  = Ccontract_w*gammaW_p
            Fcross_w_m  = Ccontract_w*gammaW_m
      
            ## First Michelson antenna patterns
            ## Calculate Fplus
            Fplus1 = (Fplus_u_p - Fplus_v_p)
            Fplus2 = (Fplus_w_p - Fplus_u_m)
            Fplus3 = (Fplus_v_m - Fplus_w_m)
    
            ## Calculate Fcross
            Fcross1 = (Fcross_u_p - Fcross_v_p)
            Fcross2 = (Fcross_w_p - Fcross_u_m)
            Fcross3 = (Fcross_v_m - Fcross_w_m)

            ## Detector response summed over polarization and integrated over sky direction
            R1[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(Fplus1))**2 + (np.absolute(Fcross1))**2, axis=1)
            R2[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(Fplus2))**2 + (np.absolute(Fcross2))**2, axis=1)
            R3[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(Fplus3))**2 + (np.absolute(Fcross3))**2, axis=1)
        
        ## Output detector response arrays; these can then be loaded instead of calculated for future analyses of the same data.
        np.savetxt('R1arrayMich.txt',R1)
        np.savetxt('R2arrayMich.txt',R2)
        np.savetxt('R3arrayMich.txt',R3)
        
        return R1, R2, R3

    def isgwb_oxyz_response(self, f0, tsegmid): 

        '''
        Calcualte the Antenna pattern/ detector transfer function functions to an isotropic SGWB using X, Y and Z TDI channels for an orbiting LISA. Note that since this is the response to an isotropic background, the response function is integrated over sky direction and averaged over polarozation. The angular integral is a linear and rectangular in the cos(theta) and phi space.  Note that f0 is (pi*L*f)/c and is input as an array

        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)
            
        tsegmid  :  array
            A numpy array of the midpoints for each time integration segment.
            
        rs1, rs2, rs3  :  array
            Satellite position vectors.

        Returns
        ---------

        R1, R2 and R3   :   float
            Antenna Patterns for the given sky direction for the three channels, integrated over sky direction and averaged over polarization, evaluated at the midpoint of each time segment.
        '''
        
        self.rs1, self.rs2, self.rs3 = self.lisa_orbits(tsegmid)

        ## Setup healpix map
        npix = hp.nside2npix(self.params['nside'])

        # Array of pixel indices
        pix_idx  = np.arange(npix)

        #Angular coordinates of pixel indcides
        theta, phi = hp.pix2ang(self.params['nside'], pix_idx)

        # Take cosine.
        ctheta = np.cos(theta)
        stheta = np.sqrt(1-ctheta**2)

        # Area of each pixel in sq.radians
        dOmega = hp.pixelfunc.nside2pixarea(self.params['nside'])

        ## Indices of midpoints array
        timeindices = np.arange(len(tsegmid))

        # Initlize arrays for the detector reponse
        R1 = np.zeros((len(timeindices),f0.size))
        R2 = np.zeros((len(timeindices),f0.size))
        R3 = np.zeros((len(timeindices),f0.size))

        ## arm vectors
        uvec = self.rs2 - self.rs1
        vvec = self.rs3 - self.rs1
        wvec = self.rs3 - self.rs2

        ## Calculate arm lengths
        Lu = np.sqrt(np.sum(uvec*uvec, axis=0))
        Lv = np.sqrt(np.sum(vvec*vvec, axis=0))
        Lw = np.sqrt(np.sum(wvec*wvec, axis=0))

        ## Define x/y/z for each satellite at all times
        x1, y1, z1 = self.rs1[0, :], self.rs1[1, :], self.rs1[2, :] 
        x2, y2, z2 = self.rs2[0, :], self.rs2[1, :], self.rs2[2, :] 
        x3, y3, z3 = self.rs3[0, :], self.rs3[1, :], self.rs3[2, :] 
        
        udir = np.tensordot((x2-x1)/Lu, np.cos(phi)*stheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z2-z1)/Lu, ctheta, axes = 0)

        vdir = np.tensordot((x3-x1)/Lv, np.cos(phi)*stheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z3-z1)/Lv, ctheta, axes = 0)

        wdir = np.tensordot((x3-x2)/Lw, np.cos(phi)*stheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z3-z2)/Lw, ctheta, axes = 0)


        ## Calculate 1/2(u x u):eplus
        Pcontract_u = 1/2*( (np.tensordot((x2-x1)/Lu, np.sin(phi), axes=0) - np.tensordot((y2-y1)/Lu, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x2-x1)/Lu, np.cos(phi)*ctheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z2-z1)/Lu, stheta, axes=0))**2  )

        Pcontract_v = 1/2*( (np.tensordot((x3-x1)/Lv, np.sin(phi), axes=0) - np.tensordot((y3-y1)/Lv, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x3-x1)/Lv, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z3-z1)/Lv, stheta, axes=0))**2  )

        Pcontract_w = 1/2*( (np.tensordot((x3-x2)/Lw, np.sin(phi), axes=0) - np.tensordot((y3-y2)/Lw, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x3-x2)/Lw, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z3-z2)/Lw, stheta, axes=0))**2  )                           


        Ccontract_u = (np.tensordot((x2-x1)/Lu, np.sin(phi), axes=0) - np.tensordot((y2-y1)/Lu, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x2-x1)/Lu, np.cos(phi)*ctheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z2-z1)/Lu, stheta, axes=0) )  

        Ccontract_v = (np.tensordot((x3-x1)/Lv, np.sin(phi), axes=0) - np.tensordot((y3-y1)/Lv, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x3-x1)/Lv, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z3-z1)/Lv, stheta, axes=0) ) 

        Ccontract_w = (np.tensordot((x3-x2)/Lw, np.sin(phi), axes=0) - np.tensordot((y3-y2)/Lw, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x3-x2)/Lw, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z3-z2)/Lw, stheta, axes=0) ) 



        # Calculate the detector response for each frequency
        for ii in range(0, f0.size):
            # Calculate GW transfer function for the michelson channels
            gammaU_p    =    1/2 * (np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(3 + udir)) + \
                                        np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(1 + udir)))
            gammaU_m    =    1/2 * (np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(3 - udir)) + \
                                        np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(1 - udir)))
                
            gammaV_p    =    1/2 * (np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + vdir)) + \
                                        np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(1+vdir)))
            gammaV_m    =    1/2 * (np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - vdir)) + \
                                        np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - vdir)))
                
            gammaW_p    =    1/2 * (np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + wdir)) + \
                                        np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 + wdir)))
            gammaW_m    =    1/2 * (np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - wdir)) + \
                                        np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - wdir)))
            
            ## Michelson Channel Antenna patterns for + pol
            ## Fplus_u = 1/2(u x u)Gamma(udir, f):eplus
            Fplus_u_p   = Pcontract_u*gammaU_p
            Fplus_u_m   = Pcontract_u*gammaU_m
            Fplus_v_p   = Pcontract_v*gammaV_p
            Fplus_v_m   = Pcontract_v*gammaV_m
            Fplus_w_p   = Pcontract_w*gammaW_p
            Fplus_w_m   = Pcontract_w*gammaW_m
    
            ## Michelson Channel Antenna patterns for x pol
            ## Fcross_u = 1/2(u x u)Gamma(udir, f):ecross
            Fcross_u_p  = Ccontract_u*gammaU_p
            Fcross_u_m  = Ccontract_u*gammaU_m
            Fcross_v_p  = Ccontract_v*gammaV_p
            Fcross_v_m  = Ccontract_v*gammaV_m
            Fcross_w_p  = Ccontract_w*gammaW_p
            Fcross_w_m  = Ccontract_w*gammaW_m
      
            ## First Michelson antenna patterns
            ## Calculate Fplus
            Fplus1 = (Fplus_u_p - Fplus_v_p)
            Fplus2 = (Fplus_w_p - Fplus_u_m)
            Fplus3 = (Fplus_v_m - Fplus_w_m)
    
            ## Calculate Fcross
            Fcross1 = (Fcross_u_p - Fcross_v_p)
            Fcross2 = (Fcross_w_p - Fcross_u_m)
            Fcross3 = (Fcross_v_m - Fcross_w_m)        

            ## Calculate antenna patterns for the X, Y, Z channels.
            FXplus = 2*np.sin(2*f0[ii])*Fplus1
            FYplus = 2*np.sin(2*f0[ii])*Fplus2
            FZplus = 2*np.sin(2*f0[ii])*Fplus3

            FXcross = 2*np.sin(2*f0[ii])*Fcross1
            FYcross = 2*np.sin(2*f0[ii])*Fcross2
            FZcross = 2*np.sin(2*f0[ii])*Fcross3


            ## Detector response for the TDI Channels, summed over polarization
            ## and integrated over sky direction
            R1[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(FXplus))**2 + (np.absolute(FXcross))**2, axis=1)
            R2[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(FYplus))**2 + (np.absolute(FYcross))**2, axis=1)
            R3[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(FZplus))**2 + (np.absolute(FZcross))**2, axis=1)

        ## Output detector response arrays; these can then be loaded instead of calculated for future analyses of the same data.
        np.savetxt('R1arrayXYZ.txt',R1)
        np.savetxt('R2arrayXYZ.txt',R2)
        np.savetxt('R3arrayXYZ.txt',R3)
        
        return R1, R2, R3

    def isgwb_oaet_response(self, f0, tsegmid): 

        '''
        Calcualte the Antenna pattern/ detector transfer function functions to an isotropic SGWB using A, E and T TDI channels. Note that since this is the response to an isotropic background, the response function is integrated over sky direction and averaged over polarization. The angular integral is a linear and rectangular in the cos(theta) and phi space.  Note that f0 is (pi*L*f)/c and is input as an array.

        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)
        
        tsegmid  :  array
            A numpy array of the midpoints for each time integration segment.
        
        rs1, rs2, rs3  :  array
            Satellite position vectors.
          
        Returns
        ---------

        R1, R2 and R3   :   arrays
            Antenna Patterns for the given sky direction for the three channels, integrated over sky direction and averaged over polarizationevaluated at the midpoint of each time segment.
        '''

        self.rs1, self.rs2, self.rs3 = self.lisa_orbits(tsegmid)

        ## Setup healpix map
        npix = hp.nside2npix(self.params['nside'])

        # Array of pixel indices
        pix_idx  = np.arange(npix)

        #Angular coordinates of pixel indcides
        theta, phi = hp.pix2ang(self.params['nside'], pix_idx)

        # Take cosine.
        ctheta = np.cos(theta)
        stheta = np.sqrt(1-ctheta**2)

        # Area of each pixel in sq.radians
        dOmega = hp.pixelfunc.nside2pixarea(self.params['nside'])

        ## Indices of midpoints array
        timeindices = np.arange(len(tsegmid))

        # Initlize arrays for the detector reponse
        R1 = np.zeros((len(timeindices),f0.size))
        R2 = np.zeros((len(timeindices),f0.size))
        R3 = np.zeros((len(timeindices),f0.size))

        ## arm vectors
        uvec = self.rs2 - self.rs1
        vvec = self.rs3 - self.rs1
        wvec = self.rs3 - self.rs2

        ## Calculate arm lengths
        Lu = np.sqrt(np.sum(uvec*uvec, axis=0))
        Lv = np.sqrt(np.sum(vvec*vvec, axis=0))
        Lw = np.sqrt(np.sum(wvec*wvec, axis=0))

        ## Define x/y/z for each satellite at all times
        x1, y1, z1 = self.rs1[0, :], self.rs1[1, :], self.rs1[2, :] 
        x2, y2, z2 = self.rs2[0, :], self.rs2[1, :], self.rs2[2, :] 
        x3, y3, z3 = self.rs3[0, :], self.rs3[1, :], self.rs3[2, :] 
        
        udir = np.tensordot((x2-x1)/Lu, np.cos(phi)*stheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z2-z1)/Lu, ctheta, axes = 0)

        vdir = np.tensordot((x3-x1)/Lv, np.cos(phi)*stheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z3-z1)/Lv, ctheta, axes = 0)

        wdir = np.tensordot((x3-x2)/Lw, np.cos(phi)*stheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*stheta, axes=0) + \
                            np.tensordot((z3-z2)/Lw, ctheta, axes = 0)


        ## Calculate 1/2(u x u):eplus
        Pcontract_u = 1/2*( (np.tensordot((x2-x1)/Lu, np.sin(phi), axes=0) - np.tensordot((y2-y1)/Lu, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x2-x1)/Lu, np.cos(phi)*ctheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z2-z1)/Lu, stheta, axes=0))**2  )

        Pcontract_v = 1/2*( (np.tensordot((x3-x1)/Lv, np.sin(phi), axes=0) - np.tensordot((y3-y1)/Lv, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x3-x1)/Lv, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z3-z1)/Lv, stheta, axes=0))**2  )

        Pcontract_w = 1/2*( (np.tensordot((x3-x2)/Lw, np.sin(phi), axes=0) - np.tensordot((y3-y2)/Lw, np.cos(phi), axes=0))**2 - \
                    (np.tensordot((x3-x2)/Lw, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*ctheta, axes=0) - \
                    np.tensordot((z3-z2)/Lw, stheta, axes=0))**2  )                           


        Ccontract_u = (np.tensordot((x2-x1)/Lu, np.sin(phi), axes=0) - np.tensordot((y2-y1)/Lu, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x2-x1)/Lu, np.cos(phi)*ctheta, axes=0) + np.tensordot((y2-y1)/Lu, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z2-z1)/Lu, stheta, axes=0) )  

        Ccontract_v = (np.tensordot((x3-x1)/Lv, np.sin(phi), axes=0) - np.tensordot((y3-y1)/Lv, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x3-x1)/Lv, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y1)/Lv, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z3-z1)/Lv, stheta, axes=0) ) 

        Ccontract_w = (np.tensordot((x3-x2)/Lw, np.sin(phi), axes=0) - np.tensordot((y3-y2)/Lw, np.cos(phi), axes=0) ) * \
                     ( np.tensordot((x3-x2)/Lw, np.cos(phi)*ctheta, axes=0) + np.tensordot((y3-y2)/Lw, np.cos(phi)*ctheta, axes=0) - \
                        np.tensordot((z3-z2)/Lw, stheta, axes=0) ) 



        # Calculate the detector response for each frequency
        for ii in range(0, f0.size):
            # Calculate GW transfer function for the michelson channels
            gammaU_p    =    1/2 * (np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(3 + udir)) + \
                                        np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(1 + udir)))
            gammaU_m    =    1/2 * (np.sinc((f0[ii])*(1 + udir)/np.pi)*np.exp(-1j*f0[ii]*(3 - udir)) + \
                                        np.sinc((f0[ii])*(1 - udir)/np.pi)*np.exp(-1j*f0[ii]*(1 - udir)))
                
            gammaV_p    =    1/2 * (np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + vdir)) + \
                                        np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(1+vdir)))
            gammaV_m    =    1/2 * (np.sinc((f0[ii])*(1 + vdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - vdir)) + \
                                        np.sinc((f0[ii])*(1 - vdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - vdir)))
                
            gammaW_p    =    1/2 * (np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 + wdir)) + \
                                        np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 + wdir)))
            gammaW_m    =    1/2 * (np.sinc((f0[ii])*(1 + wdir)/np.pi)*np.exp(-1j*f0[ii]*(3 - wdir)) + \
                                        np.sinc((f0[ii])*(1 - wdir)/np.pi)*np.exp(-1j*f0[ii]*(1 - wdir)))
            
            ## Michelson Channel Antenna patterns for + pol
            ## Fplus_u = 1/2(u x u)Gamma(udir, f):eplus
            Fplus_u_p   = Pcontract_u*gammaU_p
            Fplus_u_m   = Pcontract_u*gammaU_m
            Fplus_v_p   = Pcontract_v*gammaV_p
            Fplus_v_m   = Pcontract_v*gammaV_m
            Fplus_w_p   = Pcontract_w*gammaW_p
            Fplus_w_m   = Pcontract_w*gammaW_m
    
            ## Michelson Channel Antenna patterns for x pol
            ## Fcross_u = 1/2(u x u)Gamma(udir, f):ecross
            Fcross_u_p  = Ccontract_u*gammaU_p
            Fcross_u_m  = Ccontract_u*gammaU_m
            Fcross_v_p  = Ccontract_v*gammaV_p
            Fcross_v_m  = Ccontract_v*gammaV_m
            Fcross_w_p  = Ccontract_w*gammaW_p
            Fcross_w_m  = Ccontract_w*gammaW_m
      
            ## First Michelson antenna patterns
            ## Calculate Fplus
            Fplus1 = (Fplus_u_p - Fplus_v_p)
            Fplus2 = (Fplus_w_p - Fplus_u_m)
            Fplus3 = (Fplus_v_m - Fplus_w_m)
    
            ## Calculate Fcross
            Fcross1 = (Fcross_u_p - Fcross_v_p)
            Fcross2 = (Fcross_w_p - Fcross_u_m)
            Fcross3 = (Fcross_v_m - Fcross_w_m)        
                
    
            ## Calculate antenna patterns for the A, E and T channels -  We are switiching to doppler channel.
            FAplus = (1/3)*np.sin(2*f0[ii])*(2*Fplus1 - Fplus2 - Fplus3)
            FEplus = (1/np.sqrt(3))*np.sin(2*f0[ii])*(Fplus3 - Fplus2)
            FTplus = (1/3)*np.sin(2*f0[ii])*(Fplus1 + Fplus3 + Fplus2)
    
            FAcross = (1/3)*np.sin(2*f0[ii])*(2*Fcross1 - Fcross2 - Fcross3)
            FEcross = (1/np.sqrt(3))*np.sin(2*f0[ii])*(Fcross3 - Fcross2)
            FTcross = (1/3)*np.sin(2*f0[ii])*(Fcross1 + Fcross3 + Fcross2)
    
            ## Detector response for the TDI Channels, summed over polarization
            ## and integrated over sky direction
            R1[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(FAplus))**2 + (np.absolute(FAcross))**2, axis=1)
            R2[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(FEplus))**2 + (np.absolute(FEcross))**2, axis=1)
            R3[:, ii] = dOmega/(8*np.pi)*np.sum((np.absolute(FTplus))**2 + (np.absolute(FTcross))**2, axis=1)


        ## Output detector response arrays; these can then be loaded instead of calculated for future analyses of the same data.
        np.savetxt('R1arrayAET.txt',R1)
        np.savetxt('R2arrayAET.txt',R2)
        np.savetxt('R3arrayAET.txt',R3)
                
        return R1, R2, R3

    def tdi_aniso_sph_sgwb_response(self, f0): 

        '''
        Calculate the Antenna pattern/ detector transfer function functions to acSGWB using A, E and T TDI channels, and using a spherical harmonic decomposition. Note that the response function is integrated over sky direction with the appropriate legandre polynomial, and averaged over polarozation. Finally note that the spherical harmonic coeffcients correspond to strain sky distribution, while the legandre polynomials describe the power sky. The angular integral is a linear and rectangular in the cos(theta) and phi space.  Note that f0 is (pi*L*f)/c and is input as an array

        

        Parameters
        -----------

        f0   : float
            A numpy array of scaled frequencies (see above for def)

    

        Returns
        ---------

        R1, R2 and R3   :   float
            Antenna Patterns for the given sky direction for the three channels, integrated over sky direction and averaged over polarization. The arrays are 2-d, one direction corresponds to frequency and the other to the l coeffcient. 
        '''

        
        tt = np.arange(-1, 1, 0.02)
        pp = np.arange(0, 2*np.pi, np.pi/100)

        [ct, phi] = np.meshgrid(tt,pp)
        dct = ct[0, 1] - ct[0,0]
        dphi = phi[1,0] - phi[0,0]

        ## udir is just u.r, where r is the directional vector
        udir = np.sqrt(1-ct**2) * np.sin(phi + np.pi/6)
        vdir = np.sqrt(1-ct**2) * np.sin(phi - np.pi/6)
        wdir = vdir - udir

        # Initlize arrays for the detector reponse
        R1 = np.zeros((f0.size, self.params['lmax'] +1))
        R2 = np.zeros((f0.size, self.params['lmax'] +1))
        R3 = np.zeros((f0.size, self.params['lmax'] +1))

        ## initalize array for plms
        plms = np.zeros((tt.size, self.params['lmax']+1, self.params['lmax'] +1 ))


        ## Get associated legandre polynomials.
        for ii in range(tt.size):
            plms[ii, :, :], _ = lpmn(self.params['lmax'], self.params['lmax'], tt[ii]) 

        ## It is the squares of the polynomials which are relevent. 
        plms = plms**2
        # Calculate the detector response for each frequency
        for ii in range(0, f0.size):

            # Calculate GW transfer function for the michelson channels
            gammaU    =    1/2 * (np.sinc((f0[ii])*(1-udir))*np.exp(-1j*f0[ii]*(3+udir)) + \
                             np.sinc((f0[ii])*(1+udir))*np.exp(-1j*f0[ii]*(1+udir)))

            gammaV    =    1/2 * (np.sinc((f0[ii])*(1-vdir))*np.exp(-1j*f0[ii]*(3+vdir)) + \
                             np.sinc((f0[ii])*(1+vdir))*np.exp(-1j*f0[ii]*(1+vdir)))

            gammaW    =    1/2 * (np.sinc((f0[ii])*(1-wdir))*np.exp(-1j*f0[ii]*(3+wdir)) + \
                             np.sinc((f0[ii])*(1+wdir))*np.exp(-1j*f0[ii]*(1+wdir)))

            ## Michelson Channel Antenna patterns for + pol
            ##  Fplus_u = 1/2(u x u)Gamma(udir, f):eplus
            Fplus_u   = 1/2*(1/4*(1-ct**2) + 1/2*(ct**2)*(np.cos(phi))**2 - np.sqrt(3/16)*np.sin(2*phi)*(1+ct**2) + \
                            0.5*((np.cos(phi))**2 - ct**2))*gammaU

            Fplus_v   = 1/2*(1/4*(1-ct**2) + 1/2*(ct**2)*(np.cos(phi))**2 + np.sqrt(3/16)*np.sin(2*phi)*(1+ct**2)+ \
                         0.5*((np.cos(phi))**2 - ct**2))*gammaV

            Fplus_w   = 1/2*(1 - (1+ct**2)*(np.cos(phi))**2)*gammaW


            ## Michelson Channel Antenna patterns for x pol
            ##  Fcross_u = 1/2(u x u)Gamma(udir, f):ecross
            Fcross_u  = - np.sqrt(1-ct**2)/2 * (np.sin(2*phi + np.pi/3))*gammaU
            Fcross_v  = - np.sqrt(1-ct**2)/2 * (np.sin(2*phi - np.pi/3))*gammaV
            Fcross_w  = 1/2*ct*np.sin(2*phi)*gammaW


            ## First Michelson antenna patterns
            ## Calculate Fplus
            Fplus1 = (Fplus_u - Fplus_v)
            Fplus2 = (Fplus_w - Fplus_u)
            Fplus3 = (Fplus_v - Fplus_w)

            ## Calculate Fcross
            Fcross1 = (Fcross_u - Fcross_v)
            Fcross2 = (Fcross_w - Fcross_u)
            Fcross3 = (Fcross_v - Fcross_w)

            ## Calculate antenna patterns for the A, E and T channels -  We are switiching to doppler channel.
            FAplus = (1/3)*np.sin(2*f0[ii])*(2*Fplus1 - Fplus2 - Fplus3)
            FEplus = (1/np.sqrt(3))*np.sin(2*f0[ii])*(Fplus3 - Fplus2)
            FTplus = (1/3)*np.sin(2*f0[ii])*(Fplus1 + Fplus3 + Fplus2)

            FAcross = (1/3)*np.sin(2*f0[ii])*(2*Fcross1 - Fcross2 - Fcross3)
            FEcross = (1/np.sqrt(3))*np.sin(2*f0[ii])*(Fcross3 - Fcross2)
            FTcross = (1/3)*np.sin(2*f0[ii])*(Fcross1 + Fcross3 + Fcross2)

            ## Detector response for the TDI Channels, summed over polarization
            ## and integrated over sky direction
            
            R1[ii, :] = dct*dphi/(4*np.pi)*np.sum(np.tensordot((np.absolute(FAplus))**2 + \
                    (np.absolute(FAcross))**2, plms, axes=1), axis=(0, 1))
            R2[ii, :] = dct*dphi/(4*np.pi)*np.sum(np.tensordot((np.absolute(FEplus))**2 + \
                    (np.absolute(FEcross))**2, plms, axes=1), axis=(0, 1))
            R3[ii, :] = dct*dphi/(4*np.pi)*np.sum(np.tensordot((np.absolute(FTplus))**2 + \
                    (np.absolute(FTcross))**2, plms, axes=1), axis=(0,1))   



        return R1, R2, R3


import setuptools
from setuptools import find_packages

with open('requirements.txt', 'r') as f:
    requirements = f.readlines()
    
with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='blip-alpha',
    version='0.0.14',
    author='Sharan Banagiri',
    author_email='banag002@umn.edu',
    description='A bayesian pipeline for detecting stochastic backgrounds with LISA.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'blip = runblip.blip:main'
        ]
    },
    install_requires=requirements,
    include_package_data=True
)



import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import corner
import pickle, argparse
from healpy import Alm


def plotmaker(params, parameters, inj):

    '''
    Make posterior plots from the samples generated by tge mcmc/nested sampling algorithm.

    Parameters
    -----------

    params : dictionary
        Dictionary of config params

    parameters: string
        Array or list of strings with names of the parameters

    inj : int
        Injection dict
    '''

    post = np.loadtxt(params['out_dir'] + "/post_samples.txt")


    ## setup the truevals dict
    truevals = []

    if params['modeltype']=='isgwb':

        truevals.append(inj['log_Np'])
        truevals.append( inj['log_Na'])
        truevals.append( inj['alpha'] )
        truevals.append( inj['ln_omega0'] )

    elif params['modeltype']=='noise_only':

        truevals.append(inj['log_Np'])
        truevals.append( inj['log_Na'])

    elif params['modeltype'] =='isgwb_only':

        truevals.append( inj['alpha'] )
        truevals.append( inj['ln_omega0'] )

    elif params['modeltype']=='sph_sgwb':

        truevals.append(inj['log_Np'])
        truevals.append( inj['log_Na'])
        truevals.append( inj['alpha'] )
        truevals.append( inj['ln_omega0'] )

        ## get blms
        for lval in range(1, params['lmax'] + 1):
            for mval in range(lval + 1):

                idx = Alm.getidx(params['lmax'], lval, mval)

                if mval == 0:
                    truevals.append(np.real(inj['blms'][idx]))
                else:
                    truevals.append(np.abs(inj['blms'][idx]))
                    truevals.append(np.angle(inj['blms'][idx]))




    if len(truevals) > 0:
        knowTrue = 1 ## Bit for whether we know the true vals or not
    else:
        knowTrue = 0


    npar = len(parameters)

    plotrange = [0.999]*npar

    if params['out_dir'][-1] != '/':
        params['out_dir'] = params['out_dir'] + '/'

    ## Make corner plots
    fig = corner.corner(post, range=plotrange, labels=parameters, quantiles=(0.16, 0.84),
                        smooth=None, smooth1d=None, show_titles=True,
                        title_kwargs={"fontsize": 12},label_kwargs={"fontsize": 14},
                        fill_contours=True, use_math_text=True, )


    # Put correct values
    # Extract the axes
    axes = np.array(fig.axes).reshape((npar, npar))
    for ii in range(npar):
        ax = axes[ii, ii]

        ## Draw truevals if they exist
        if knowTrue:
            ax.axvline(truevals[ii], color="g", label='true value')

    ## Save posterior
    plt.savefig(params['out_dir'] + 'corners.png', dpi=150)
    print("Posteriors plots printed in " + params['out_dir'] + "corners.png")
    plt.close()

if __name__ == '__main__':

    # Create parser
    parser = argparse.ArgumentParser(prog='plotmaker', usage='%(prog)s [options] rundir', description='run plotmaker')

    # Add arguments
    parser.add_argument('rundir', metavar='rundir', type=str, help='The path to the run directory')

    # execute parser
    args = parser.parse_args()


    paramfile = open(args.rundir + '/config.picle')

    ## things are loaded from the pickle file in the same order they are put in
    params = pickle.load(paramfile)
    inj = pickle.load(paramfile)
    parameters = pickle.load(paramfile)

    plotmaker(params, parameters, inj)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib
#mpl.rcParams['figure.figsize'] = (14,10)
matplotlib.rcParams.update(matplotlib.rcParamsDefault)

def lisaPSD(L=2.5e9, channel='TDI', fmin=5e-6, fmax=1e0, delf=1e-6, doPlot=False):
    '''
    This script calcualtes the power spectrum of LISA channels. It
    assumes an equal arm staionary LISA. If channel flag is 'M', then the
    michelson channels are given.


    If the channels flag is TDI, we implement TDI as
    described in http://iopscience.iop.org/article/10.1088/0264-9381/18/17/308

    The noise levels are taken from the 2017 LISA proposal. The other input is
    the arm length of IFO, the default is 2.5 million km

    '''

    # speed of light
    cspeed = 3e8 #m/s

    frange = np.arange(fmin, fmax, delf) # in Hz

    numfreqs = frange.size

    #Charactersitic frequency
    fstar = cspeed/(2*np.pi*L)

    # define f0 = f/2f*
    f0 = frange/(2*fstar)


    Tyear = 365.24*24*60*60     # in seconds

    H0 = 2.2*10**(-18)         # in SI units, from the planck observations


    # 2017 Lisa proposal noise estimations

    # Position noise, converted to phase
    Sp = 4e-42*(1 + ((2e-3)/frange)**4)
    # Acceleration noise converted to phase
    Sa = 3.6e-49*(1+ ((4e-4)/(frange))**2)*(1+ (frange/(8e-3))**4)*(1/(2*np.pi*frange)**4)

    ###############################
    #   Old values from Adams and Cornish, 2010
    #
    #   Sp = 4e-42*(1+ 0*frange)
    #   Sa = 9e-50*(1+ ((1e-4)/(frange))**2)*(1/(2*np.pi*frange)**4)
    #
    ###################################


    ## Noise spectra of the TDI Channels
    if channel == 'TDI':
        # A channel
        S1  = (4/9)*(np.sin(2*f0))**2*(np.cos(2*f0)*(12*Sp) + 24*Sp )+ \
        (16/9)*(np.sin(2*f0))**2*(np.cos(2*f0)*12*Sa + np.cos(4*f0)*(6*Sa) + 18*Sa)
        # E  channel
        S2  = (4/3)*(np.sin(2*f0))**2*(4*Sp + (4 + 4*np.cos(2*f0))*Sp) +\
        (16/3)*(np.sin(2*f0))**2*(4*Sa + 4*Sa*np.cos(2*f0) +  4*Sa*(np.cos(2*f0))**2 )
        # T channel
        S3 = (4/9)*(np.sin(2*f0))**2* (12 - 12*np.cos(2*f0))*Sp + \
        (16/9)*(np.sin(2*f0))**2*(6 - 12*np.cos(2*f0) + 6*(np.cos(2*f0))**2)*Sa

        np.save('LISA_2017_PSD_TDI', (frange, S1,S2, S3 ))

        if doPlot:
            plt.plot(frange, S1, label = 'Channel A')
            plt.plot(frange, S2, label = 'Channel E')
            plt.plot(frange, S3, label = 'Channel T')
            plt.legend()
            plt.xlabel('Frequency')
            plt.ylabel('Noise spectrum sqrt(Hz) ')
            plt.gca().set_xlim(5*1e-4, 0.1)
            plt.gca().set_ylim(1e-49, 1e-38)
            plt.xscale('log')
            plt.yscale('log')
            plt.savefig('TDI_PSD.png', dpi=150)
            plt.close()




    elif channel == 'M': #Michelson Channels

        S1 = 4*Sp + 8*Sa*(1 + (np.cos(2*f0))**2)
        S2 = S1
        S3 = S1

        np.save('LISA_2017_PSD_M', (frange, S1,S2, S3 ))

        if doPlot:
            plt.plot(frange, S1, label = 'Michelson channel')
            plt.legend()
            plt.xlabel('Frequency')
            plt.ylabel('Noise spectrum sqrt(Hz) ')
            plt.gca().set_xlim(5*1e-4, 0.1)
            plt.gca().set_ylim(1e-43, 1e-37)
            plt.xscale('log')
            plt.yscale('log')
            plt.savefig('Mchannel_PSD.png', dpi=150)
            plt.close()



import numpy as np
from dynesty import NestedSampler
from dynesty.utils import resample_equal


class dynesty_engine():

    '''
    Class for interfacing with dynesty sampler. This method also contains the
    priors definition for all models written to work with the dynesty sampler.
    '''


    @classmethod
    def define_engine(cls, lisaobj, params, nlive, randst):


        # create the nested sampler objects
        if params['modeltype']=='isgwb':

            print("Doing an isotropic stochastic analysis...")
            parameters = [r'$\log_{10} (Np)$', r'$\log_{10} (Na)$', r'$\alpha$', r'$\log_{10} (\Omega_0)$']
            npar = len(parameters)

            engine = NestedSampler(lisaobj.isgwb_log_likelihood, cls.isgwb_prior,\
                    npar, bound='multi', sample='rwalk', nlive=nlive, rstate = randst)

        elif params['modeltype']=='sph_sgwb':

            print("Doing a spherical harmonic stochastic analysis ...")

            # add the basic parameters first
            parameters = [r'$\log_{10} (Np)$', r'$\log_{10} (Na)$', r'$\alpha$', r'$\log_{10} (\Omega_0)$']

            # list for imposing periodic boundary conditions on phase variables
            periodic_bc = []        

            # add the blms
            for lval in range(1, params['lmax'] + 1):
                for mval in range(lval + 1):

                    if mval == 0:
                        parameters.append(r'$b_{' + str(lval) + str(mval) + '}$' )
                    else:
                        parameters.append(r'$|b_{' + str(lval) + str(mval) + '}|$' )
                        parameters.append(r'$\phi_{' + str(lval) + str(mval) + '}$' )
                        
                        # keep track of phase variable positions
                        periodic_bc.append(len(parameters) - 1)

            ## RM is line later.
            # parameters.append(r'$|b_{' + str(1) + str(1) + '}|$' )
            # parameters.append(r'$\phi_{' + str(1) + str(1) + '}$' )
            npar = len(parameters)

            engine = NestedSampler(lisaobj.sph_log_likelihood, cls.sph_prior, \
                    npar, bound='multi', sample='rwalk', nlive=nlive, rstate = randst, periodic=periodic_bc)

        elif params['modeltype']=='noise_only':

            print("Doing an instrumental noise only analysis ...")
            parameters = [r'$\log_{10} (Np)$', r'$\log_{10} (Na)$']
            npar = len(parameters)

            engine = NestedSampler(lisaobj.instr_log_likelihood,  cls.instr_prior,\
                    npar, bound='multi', sample='rwalk', nlive=nlive, rstate = randst)

        elif params['modeltype'] =='isgwb_only':

            print("Doing an isgwb signal only analysis ...")
            parameters = [r'$\alpha$', r'$\log_{10} (\Omega_0)$']
            npar = len(parameters)

            engine = NestedSampler(lisaobj.isgwb_only_log_likelihood, cls.isgwb_only_prior,\
                    npar, bound='multi', sample='rwalk', nlive=nlive, rstate = randst)

        else:
            raise ValueError('Unknown recovery model selected')

        # print npar
        print("npar = " + str(npar))

        return engine, parameters

    @staticmethod
    def run_engine(engine):


       # -------------------- Run nested sampler ---------------------------
        engine.run_nested(dlogz=0.5,print_progress=True )

        # re-scale weights to have a maximum of one
        res = engine.results
        weights = np.exp(res['logwt'] - res['logz'][-1])
        weights[-1] = 1 - np.sum(weights[0:-1])

        post_samples = resample_equal(res.samples, weights)

        # Pull the evidence and the evidence error
        logz = res['logz']
        logzerr = res['logzerr']


        return post_samples, logz, logzerr

    @staticmethod
    def instr_prior(theta):


        '''
        Prior function for only instrumental noise

        Parameters
        -----------

        theta   : float
            A list or numpy array containing samples from a unit cube.

        Returns
        ---------

        theta   :   float
            theta with each element rescaled. The elements are  interpreted as alpha, omega_ref, Np and Na

        '''


        # Unpack: Theta is defined in the unit cube
        log_Np, log_Na = theta

        # Transform to actual priors
        log_Np = -5*log_Np - 39
        log_Na = -5*log_Na - 46

        return (log_Np, log_Na)

    @staticmethod
    def isgwb_prior(theta):


        '''
        Prior function for an isotropic stochastic backgound analysis.

        Parameters
        -----------

        theta   : float
            A list or numpy array containing samples from a unit cube.

        Returns
        ---------

        theta   :   float
            theta with each element rescaled. The elements are  interpreted as alpha, omega_ref, Np and Na

        '''


        # Unpack: Theta is defined in the unit cube
        log_Np, log_Na, alpha, log_omega0 = theta

        # Transform to actual priors
        alpha       =  10*alpha-5
        log_omega0  = -10*log_omega0 - 4
        log_Np      = -5*log_Np - 39
        log_Na      = -5*log_Na - 46

        return (log_Np, log_Na, alpha, log_omega0)

    @staticmethod
    def sph_prior(theta):

        '''
        Prior for a power spectra based spherical harmonic anisotropic analysis

        Parameters
        -----------

        theta   : float
            A list or numpy array containing samples from a unit cube.

        Returns
        ---------

        theta   :   float
            theta with each element rescaled. The elements are  interpreted as alpha, omega_ref for each of the harmonics, Np and Na. The first element is always alpha and the last two are always Np and Na
        '''

        # The first two are the priors on the position and acc noise terms.
        log_Np = -4*theta[0] - 39
        log_Na = -4*theta[1] - 46

        # Prior on alpha, and omega_0
        alpha = 8*theta[2] - 4
        log_omega0  = -6*theta[3] - 5

        # Calculate lmax from the size of theta blm arrays. The shape is
        # given by size = (lmax + 1)**2 - 1. The '-1' is because b00 is
        # an independent parameter
        lmax = np.sqrt( theta[4:].size + 1 ) - 1

        if lmax.is_integer():
            lmax = int(lmax)
        else:
            raise ValueError('Illegitimate theta size passed to the spherical harmonic prior')

        # The rest of the priors define the blm parameter space
        blm_theta = []

        ## counter for the rest of theta
        cnt = 4

        for lval in range(1, lmax + 1):
            for mval in range(lval + 1):

                if mval == 0:
                    blm_theta.append(6*theta[cnt] - 3)
                    cnt = cnt + 1
                else:

                    # prior on real and imaginary parts
                    # blm_theta.append(6*theta[cnt] - 3)
                    # blm_theta.append(6*theta[cnt + 1] - 3)

                    ## prior on amplitude, phase
                    blm_theta.append(3* theta[cnt])
                    blm_theta.append(2*np.pi*theta[cnt+1] - np.pi)
                    cnt = cnt + 2

        # rm these three lines later.
        # blm_theta.append(theta[4])
        # blm_theta.append(2*np.pi*theta[5] - np.pi)

        theta = [log_Np, log_Na, alpha, log_omega0] + blm_theta

        return theta

    @staticmethod
    def isgwb_only_prior(self, theta):


        '''
        Prior function for an isotropic stochastic backgound analysis.

        Parameters
        -----------

        theta   : float
            A list or numpy array containing samples from a unit cube.

        Returns
        ---------

        theta   :   float
            theta with each element rescaled. The elements are  interpreted as alpha, omega_ref

        '''

        # Unpack: Theta is defined in the unit cube
        alpha, log_omega0  = theta

        # Transform to actual priors
        alpha       = 10*alpha-5
        log_omega0  = -10*log_omega0 - 4

        return (alpha, log_omega0)
















